#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

constexpr int BLOCK_SIZE = 256;

__device__ inline float scale_value(float val, float factor) {
    return val * factor;
}

__device__ inline float hard_tanh(float val, float min_val, float max_val) {
    return fminf(fmaxf(val, min_val), max_val);
}

__device__ inline float gelu_activation(float val) {
    constexpr float c = 0.044715;
    constexpr float sqrt_2_over_pi = 0.7978845608028654;
    float x_cube = val * val * val;
    float tanh_arg = sqrt_2_over_pi * (val + c * x_cube);
    float tanh_res = tanh(tanh_arg);
    return 0.5 * val * (1.0 + tanh_res);
}

__global__ void fused_activation_with_shared_memory(
    float* __restrict__ x,
    const float scaling_factor,
    const float hardtanh_min,
    const float hardtanh_max,
    const int64_t numel) {
    __shared__ float shared_data[BLOCK_SIZE];

    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int stride = gridDim.x * blockDim.x;

    for (int i = idx; i < numel; i += stride) {
        // Load data into shared memory
        float val = x[i];

        // Use shared memory for subsequent operations
        shared_data[threadIdx.x] = scale_value(val, scaling_factor);
        shared_data[threadIdx.x] = hard_tanh(
            shared_data[threadIdx.x], hardtanh_min, hardtanh_max);
        val = gelu_activation(shared_data[threadIdx.x]);

        // Write back the computed value
        x[i] = val;
    }
}

void fused_activation_cuda(
    torch::Tensor& x,
    double scaling_factor,
    double hardtanh_min,
    double hardtanh_max) {

    const auto numel = x.numel();
    const dim3 blocks((numel + BLOCK_SIZE - 1) / BLOCK_SIZE);

    AT_DISPATCH_FLOATING_TYPES(x.scalar_type(), "fused_activation_cuda", ([&] {
        fused_activation_with_shared_memory<<<blocks, BLOCK_SIZE>>>(
            x.data_ptr<float>(),
            static_cast<float>(scaling_factor),
            static_cast<float>(hardtanh_min),
            static_cast<float>(hardtanh_max),
            numel);
    }));
}

torch::Tensor module_fn_forward(
    torch::Tensor x,
    double scaling_factor,
    double hardtanh_min,
    double hardtanh_max,
    torch::Tensor weight,
    torch::Tensor bias) {

    x = x.contiguous().cuda();
    weight = weight.contiguous().cuda();
    bias = bias.contiguous().cuda();

    auto xw = torch::matmul(x, weight.t()) + bias;
    fused_activation_cuda(xw, scaling_factor, hardtanh_min, hardtanh_max);

    return xw;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn_forward, "Module function forward (CUDA)");
}
